from Globals import InitializeClass
from AccessControl import ClassSecurityInfo
from Products.CMFDefault.DublinCore import DefaultDublinCoreImpl
from Products.CMFCore.PortalContent import PortalContent
from Products.CMFCore import CMFCorePermissions
from OFS.SimpleItem import SimpleItem
from Products.CMFCore.utils import getToolByName

no_pil=0
try: import PIL
except: no_pil=1

factory_type_information = ( { 'id': 'Questionnaire'
                             , 'content_icon': 'icon_cmfquestions_small.gif'
                             , 'meta_type': 'Questionnaire'
                             , 'description': ('You can add questions to questionnaires')
                             , 'product': 'CMFQuestions'
                             , 'factory': 'addQuestionnaire'
                             , 'immediate_view': 'questionnaire_edit_form'
                             , 'actions': ( { 'id': 'view'
                                            , 'name': 'View'
                                            , 'action': 'questionnaire_view'
                                            , 'permissions': (CMFCorePermissions.View,) 
					    }                                      
                                          , { 'id': 'edit'
                                            , 'name': 'Edit'
                                            , 'action': 'questionnaire_edit_form'
                                            , 'permissions': (CMFCorePermissions.ModifyPortalContent,)
					    }
                                          , { 'id': 'properties'
                                            , 'name': 'Properties'
                                            , 'action': 'questionnaire_edit_properties_form'
                                            , 'permissions': (CMFCorePermissions.ModifyPortalContent,)
					    } 
                                          , { 'id': 'results'
                                            , 'name': 'Review results'
                                            , 'action': 'questionnaire_view_results'
                                            , 'permissions': (CMFCorePermissions.ModifyPortalContent,)
					    } ), 
                             }, )

def addQuestionnaire(self
                     , id
                     , title=''
                     , REQUEST=None):
    """ question factory """
    o = Questionnaire( id
                       , title
                       )
    self._setObject(id, o)

class Questionnaire(PortalContent, DefaultDublinCoreImpl):
    meta_type='Questionnaire'
    security = ClassSecurityInfo()
    def __init__(self,
                 id,
                 title=''):
        """ constructor"""
        DefaultDublinCoreImpl.__init__(self)
        self.id=id
        self.questions = {}
        self.completed = []
        self.thankyou_message = "Thank you for completing the questionnaire"
        self.saved_message = """You have saved the questionnaire.
Don't forget to come back and finish it."""
        self.status = "open" # closed or open
	#self.allow_anonymous = 0
	#self.exit_url = "/" # getToolByName(self, 'portal_url').getPortalObject().absolute_url()
	#self.survey_id_no = 0

    security.declareProtected(CMFCorePermissions.View, 'Title')
    def Title(self):
        """Override DC Title to include its status: open or closed"""
        return "%s (%s)" % (self.title, getattr(self, 'status', ''))

    def getSavedMessage(self):
        if not hasattr(self, 'saved_message'):
            self.saved_message = """You have saved the questionnaire.
Don't forget to come back and finish it."""
        return self.saved_message            

    def getAllowAnonymous(self):
	if not hasattr(self, 'allow_anonymous'):
	    self.allow_anonymous = 0
	return self.allow_anonymous

    def getExitUrl(self):
	if not hasattr(self, 'exit_url'):
	    self.exit_url = "/" #getToolByName(self, 'portal_url').getPortalObject().absolute_url()
	return self.exit_url

    def getAnonymousId(self):
	# returns the id to use for an anonymous user
	portal_membership = getToolByName(self, 'portal_membership')
	if portal_membership.isAnonymousUser() and self.getAllowAnonymous():
  	  if not hasattr(self, 'survey_id_no'):
	    self.survey_id_no = 0
	  self.survey_id_no += 1
	  return 'anonymous' + str(self.survey_id_no)

	return portal_membership.getAuthenticatedMember().getId()
	
    
    def title_or_id(self):
        if self.title:
            return self.Title()
        else:
            return self.getId()
        
    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'editQuestionnaire')
    def editQuestionnaire(self,
                          thankyou_message="",
                          saved_message="",
                          status="",
                          title="",
                          description="",
			  allow_anonymous=0,
			  exit_url="/"):
        """Change the questionnaire properties"""
        self.thankyou_message = thankyou_message
        self.saved_message = saved_message
        self.status = status
        self.title = title
        self.description = description
	self.allow_anonymous = allow_anonymous
        self.exit_url = exit_url
        self.reindexObject()
    
    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'addQuestion') 
    def addQuestion(self, id=""):
        """Add a question, and return it"""
        position = len(self.questions) + 1
        if not id:
            id = str(position)
        if id in self.questions.keys():
            raise QuestionnaireError("That question already exists")
        question = Question(id, position, self)
        self.questions[id] = question
        self._p_changed = 1
        return self.questions[id]

    security.declareProtected(CMFCorePermissions.View, 'getQuestion') 
    def getQuestion(self, id):
        """Return a specific question"""
        return self.questions.get(id, '')

    security.declareProtected(CMFCorePermissions.View, 'getAllQuestions') 
    def getAllQuestions(self):
        """Return the entire list of questions, sorted by position"""
        qs = self.questions.values()
        qs.sort(lambda x, y: cmp(x.position, y.position))
        return qs

    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'checkQuestionOrder')
    def checkQuestionOrder(self):
        # Make sure question 1 is in position 1, question 2 in position 2, etc
        keys = [int(key) for key in self.questions.keys()]
        keys.sort()
        pos = 1
        for key in keys:
            if key != pos or pos != self.questions[str(key)].position:
                new_key = str(pos)
                self.questions[new_key] = self.questions[str(key)]
                self.questions[new_key].setId(new_key)
                self.questions[new_key].setPosition(pos)
		if str(key) != new_key:
                    del self.questions[str(key)]
            pos += 1
        self._p_changed = 1

    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'swapQuestions')
    def swapQuestions(self, id1, id2):
	"""Swap question order, doesn't seem to work properly, needs some fixes"""
	if not self.questions.has_key(id1) or not self.questions.has_key(id2):
	    raise StandardError, "That question does not exist"
	temp = {}
	temp[id2] = self.questions[id1]
	temp[id1] = self.questions[id2]
	temp[id2].setPosition(int(id2))
	temp[id2].setId(id2)
	temp[id1].setPosition(int(id1))
	temp[id1].setId(id1)
	self.questions.update(temp)
	del temp
	self.checkQuestionOrder()

    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'removeQuestion') 
    def removeQuestion(self, ids):
        """Remove questions"""
        for id in ids:
            if not self.questions.has_key(id):
                raise StandardError, "That question does not exist"
            del self.questions[id]
	self.checkQuestionOrder()

    def resetQuestion(self, id):
        """Remove question answers"""
        if not self.questions.has_key(id):
            raise StandardError, "That question does not exist"
        self.questions[id].reset()

    def resetAllQuestions(self):
        """Remove all question answers"""
        for q in self.questions.values():
            q.reset()
        self.completed = []
        self.reindexObject()

    def getRespondents(self):
        """Return a list of respondents"""
        qs = self.questions.values()
        users = {}
        for q in qs:
            for u in q.values_by_user.keys():
                users[u] = 1
        return users.keys()
        
    security.declareProtected(CMFCorePermissions.View, 'getQuestionsCount') 
    def getQuestionsCount(self):
        """Return a count of questions asked"""
        return len(self.questions)

    def setCompletedFor(self, userid):
        """Mark the questionnaire as completed for the user"""
        self.completed.append(userid)
        for q in self.getAllQuestions():
	    if q.getQuestionType() != 'text' and q.getQuestionType() != 'area':
                values = q.getAnswerFor(userid).split('##',len(q.getAnswerOptions()))
	        for value in values:
                    if not q.value_counts.has_key(value):
                        q.value_counts[value] = 0
                    q.value_counts[value] += 1
        q._p_changed = 1
        self._p_changed = 1

    def getCompletedFor(self, userid):
        """Has the used completed the questionnaire?"""
        return userid in self.completed

    def SearchableText(self):
        """The queried text for searches: title, description, questions"""
        questions = [x.getTitle() for x in self.getAllQuestions()]
        questions = " ".join(questions)
        return "%s %s %s" % (self.title, self.Description(), questions)

class Question(SimpleItem):
    """A question can only have one answer per id"""
    security = ClassSecurityInfo()
    def __init__(self, id, position, parent):
        self.id = id
        self.setPosition(position)

        # for querying a particular user's answers
        self.values_by_user = {}

        # for constructing histograms
        self.value_counts = {}

        # just the answer values themselves
        self.values = []

        self.parent = parent

    def reset(self):
        """Remove all my answers"""
        self.values_by_user = {}
        self.value_counts = {}
        self.values = []
        self._p_changed = 1

    def getId(self):
        """Return id"""
        return self.id

    def setId(self, id):
        """Set id"""
        self.id = id

    def setPosition(self, position):
        """Set the question position"""
        self.position = position

    def setTitle(self, title):
        """Set the question title"""
        self.title = title

    def setEnablingObjective(self, eo):
        """Set the enabling objective"""
        self.enabling_objective = eo

    def getEnablingObjective(self):
        """Get the enabling objective"""
        return self.enabling_objective

    def setDescription(self, description):
        """Set the question description"""
        self.description = description

    def setAnswerOptions(self, options):
        """Set a list of possible options for the answer"""
        if type(options) == type([]):
            self.options = options
            
    def getAnswerOptions(self):
        """get the possible answers"""
        return self.options

    def getTitle(self):
        """Return the title"""
        return self.title

    def getDescription(self):
        """Return the description"""
        return self.description

    def getShowCommentField(self):
        """Return whether to show the comment field"""
        return getattr(self, 'showComment', 1)

    def setShowCommentField(self, show):
        """Set whether to show a comment field"""
        self.showComment = show

    def getCommentLabel(self):
        """Return the comment label"""
        return getattr(self, 'commentLabel', 'Comment - mandatory if "no"')

    def setCommentLabel(self, label):
        """Set the comment label"""
        if label:
            self.commentLabel = label

    def setCommentType(self, type):
        if type:
            self.commentType = type

    def setQuestionType(self, type):
        if type:
            self.questionType = type

    def setOptional(self, optional):
	"""Set whether a question is optional"""
	self.optional = optional

    #security.declareProtected('Answer questions', 'addAnswer') 
    def addAnswer(self, userid, value, comments=""):
        """Add an answer and optional comments for a user."""
        answered = self.values_by_user.has_key(userid)
        self.values_by_user[userid] = {'value':value,
                                       'comments':comments}
        #if not answered:
        #    if not self.value_counts.has_key(value):
        #        self.value_counts[value] = 0
        #    self.value_counts[value] += 1
        self._p_changed = 1

    def getAnswerFor(self, userid):
        """Get a specific user's answer"""
        return self.values_by_user.get(userid, {}).get('value', None)

    def getCommentsFor(self, userid):
        """Get a specific user's comments"""
        return self.values_by_user.get(userid, {}).get('comments', None)

    def getAggregateAnswers(self):
        """Return a mapping of aggregrate answer values,
        suitable for a histogram"""
        return self.value_counts

    def getComments(self):
        """Return a userid, comments mapping"""
        mlist = []
        for k, v in self.values_by_user.items():
            mapping = {}
            mapping['userid'] = k
            mapping['comments'] = v.get('comments', '')
            mlist.append(mapping)
        return mlist

    def getCommentType(self):
        try:
            return self.commentType
        except:
            return "area"

    def getQuestionType(self):
        try:
            return self.questionType
        except:
            return "radio"

    def getOptional(self):
	try:
	    return getattr(self, 'optional', 0)
	except:
	    return 0

    def getAverageAnswer(self, raise_exception="no"):
        """Return an average of numeric answers.  If *raise_exception*
        is passed, then this method will cause an error if any answers
        are not numeric"""
        raise "NotImplemented"

    def getHistogram(self):
        """Return an image representing answer data"""
        raise "NotImplemented"

class QuestionnaireError(Exception):

    def __init__(self, text):
        Exception.__init__(self, text)
        self.text = text
    
InitializeClass(Questionnaire)
